#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>

#define LOOPS   100000
#define USLEEP  100

int
main(int argc, char **argv)
{
	struct sockaddr_in sin;
	int listenfd, connfd, pid;
	u_int counter, ncpus;
	size_t len;
	socklen_t solen;

	len = sizeof(ncpus);
	if (sysctlbyname("kern.smp.cpus", &ncpus, &len, NULL, 0) < 0)
		err(1, "kern.smp.cpus");
	if (len != sizeof(ncpus))
		errx(1, "kern.smp.cpus: invalid length");
	if (ncpus < 2)
		warnx("SMP not present, test may be unable to trigger race");

	/*
	 * Create a tcp socket that the child will repeatedly accept()
	 * from, and that the parent will repeatedly connect() to.
	 * Choose the listen port dynamically and get its value using
	 * getsockname().
	 */
	if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		err(1, "parent: socket error");
	memset(&sin, 0, sizeof(sin));
	sin.sin_len = sizeof(sin);
	sin.sin_family = AF_INET;
	if (bind(listenfd, (struct sockaddr *) &sin,
		 sizeof(sin)) < 0)
		err(1, "parent: bind error");
	if (getsockname(listenfd, (struct sockaddr *) &sin,
			&solen) < 0)
		err(1, "parent: getsockname error");
	if (listen(listenfd, 1024) < 0)
		err(1, "parent: listen error");
	pid = fork();
	if (pid == -1)
		err(1, "fork()");
	if (pid != 0) {
		/*
		 * In the parent, repeatedly connect and disconnect
		 * from the socket, attempting to induce the race.
		 */
		close(listenfd);
		sleep(1);
		for (counter = 0; counter < LOOPS; counter++) {
			if ((connfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
				(void)kill(pid, SIGTERM);
				err(1, "parent: socket error");
			}
			if (connect(connfd, (struct sockaddr *)&sin,
				    sizeof(sin)) < 0) {
				err(1, "parent: connect error (after %d iterations)",
				    counter);
			}
			if (close(connfd) < 0) {
				(void)kill(pid, SIGTERM);
				err(1, "parent: close error");
			}
			usleep(USLEEP);
		}
		(void)kill(pid, SIGTERM);
	} else {
		/*
		 * In the child, loop accepting, shutdown and closing.
		 * We may pick up the race here so report errors from
		 * close().
		 */
		for ( ; ; ) {
			if ((connfd = accept(listenfd,
			    (struct sockaddr *)NULL, NULL)) < 0)
				err(1, "child: accept error");
			if (shutdown(connfd, SHUT_WR) < 0)
				err(1, "child: shutdown error");
			if (close(connfd) < 0)
				err(1, "child: close error");
		}
	}
	printf("OK\n");
	exit(0);
}